#include "fw/base.h"
#include "fw/dynamixel/base.h"
#include "fw/dynamixel/protocol.h"
#include <emlabcpp/static_circular_buffer.h>

#pragma once

namespace schpin
{

// TODO: random constant much
using dynmx_reg_set_callback =
    std::add_pointer_t< std::optional< protocol_error_record >( void*, const dynmx_buffer& ) >;
using dynmx_reg_get_callback = std::add_pointer_t< dynmx_buffer( void* ) >;

struct dynmx_reg_token
{
        void*                  reg_ptr;
        dynmx_addr             addr;
        dynmx_len              size;
        dynmx_reg_set_callback set_cb_ptr;
        dynmx_reg_get_callback get_cb_ptr;

        std::optional< protocol_error_record > set_cb( const dynmx_buffer& v )
        {
                return set_cb_ptr( reg_ptr, v );
        }

        dynmx_buffer get_cb()
        {
                return get_cb_ptr( reg_ptr );
        }
};

template < dynmx_addr addr, typename T >
class dynmx_reg
{
        using prot_item = protocol_item< T >;

        T val_;

        static std::optional< protocol_error_record > set_cb( void* obj_raw, const dynmx_buffer& v )
        {
                dynmx_reg*            obj = reinterpret_cast< dynmx_reg< addr, T >* >( obj_raw );
                protocol_error_record res;
                prot_item::deserialize( v,
                                        match(
                                            [&]( value_tag, T val ) {
                                                    obj->val_ = val;
                                            },
                                            [&]( error_tag, auto e ) {
                                                    res = e;
                                            } ) );
                return res;
        }
        static dynmx_buffer get_cb( void* obj_raw )
        {
                dynmx_reg*   obj = reinterpret_cast< dynmx_reg< addr, T >* >( obj_raw );
                dynmx_buffer buffer;
                prot_item::serializer_at( buffer, obj->val_ );
                return buffer;
        }

      public:
        dynmx_reg_token get_token()
        {
                return { this, addr, prot_item::size, set_cb, get_cb };
        }

        operator const T&() const
        {
                return val_;
        }
};

struct dynmx_state
{
        // EEPROM
        dynmx_reg< 0, dynmx_model_number > model_number;
        dynmx_reg< 2, dynmx_model_info >   model_info;
        dynmx_reg< 6, dynmx_fw_version >   fw_version;
        dynmx_reg< 7, dynmx_servo_id >     id;
        dynmx_reg< 8, dynmx_baud_rate >    baud_rate;
        dynmx_reg< 9, dynmx_delay_time >   delay;
        dynmx_reg< 10, dynmx_drive_mode >  drive_mode;
        dynmx_reg< 11, dynmx_oper_mode >   oper_mode;
        dynmx_reg< 12, dynmx_servo_id >    shadow_id;
        dynmx_reg< 13, dynmx_proto_type >  proto_type;
        dynmx_reg< 20, dynmx_home_offset > home_offset;
        dynmx_reg< 24, dynmx_vel_limit >   moving_treshold;
        dynmx_reg< 31, dynmx_temp >        temp_limit;
        dynmx_reg< 32, dynmx_voltage >     max_volt;
        dynmx_reg< 34, dynmx_voltage >     min_volt;
        dynmx_reg< 36, dynmx_pwm_limit >   pwm_limit;
        dynmx_reg< 44, dynmx_vel_limit >   vel_limit;
        dynmx_reg< 48, dynmx_pos >         max_pos;
        dynmx_reg< 52, dynmx_pos >         min_pos;
        dynmx_reg< 63, dynmx_error >       shutdown;
        // RAM
        dynmx_reg< 64, dynmx_bool >           torque_en;
        dynmx_reg< 65, dynmx_bool >           led_en;
        dynmx_reg< 68, dynmx_status_level >   status_return;
        dynmx_reg< 69, dynmx_bool >           registered_inst;
        dynmx_reg< 70, dynmx_error >          error_status;
        dynmx_reg< 76, dynmx_pid_gain >       vel_i_gain;
        dynmx_reg< 78, dynmx_pid_gain >       vel_p_gain;
        dynmx_reg< 80, dynmx_pid_gain >       pos_d_gain;
        dynmx_reg< 82, dynmx_pid_gain >       pos_i_gain;
        dynmx_reg< 84, dynmx_pid_gain >       pos_p_gain;
        dynmx_reg< 88, dynmx_pid_gain >       ff_scnd_gain;
        dynmx_reg< 90, dynmx_pid_gain >       ff_frst_gain;
        dynmx_reg< 98, dynmx_watchdog >       bus_watchdog;
        dynmx_reg< 100, dynmx_pwm >           goal_pwm;
        dynmx_reg< 104, dynmx_vel >           goal_vel;
        dynmx_reg< 108, dynmx_profile_accel > profile_accel;
        dynmx_reg< 112, dynmx_profile_vel >   profile_vel;
        dynmx_reg< 116, dynmx_pos >           goal_pos;
        dynmx_reg< 120, dynmx_time >          realtime_tick;
        dynmx_reg< 122, dynmx_bool >          moving;
        dynmx_reg< 123, dynmx_status >        status;
        dynmx_reg< 124, dynmx_pwm >           present_pwm;
        dynmx_reg< 126, dynmx_load >          present_load;
        dynmx_reg< 128, dynmx_vel >           present_vel;
        dynmx_reg< 132, dynmx_pos >           present_pos;
        dynmx_reg< 136, dynmx_vel >           vel_trajectory;
        dynmx_reg< 140, dynmx_pos >           pos_trajectory;
        dynmx_reg< 144, dynmx_voltage >       present_voltage;
        dynmx_reg< 146, dynmx_temp >          present_temp;
};

class dynmx_context
{
        dynmx_state state_;

        struct buff_item
        {
                s_ctl_id           msg_id;
                dynmx_message_enum id;
                dynmx_reg_token    tok;
        };

        em::static_circular_buffer< buff_item, 16 > context_;
        s_ctl_id                                    msg_id_counter_ = 0;
        dynmx_servo_id                              id_;

      public:
        dynmx_context( dynmx_servo_id id )
          : id_( id )
        {
        }

        const dynmx_servo_id& get_id() const
        {
                return id_;
        }

        const dynmx_state& operator*() const
        {
                return state_;
        }

        const dynmx_state* operator->() const
        {
                return &state_;
        }

        void on_servo_receive( s_ctl_id msg_id, const dynmx_master_tuple& tpl )
        {
                auto [header, servo_id, status_const, error, buffer, crc] = tpl;

                auto iter = em::find_if( context_, [&]( const auto& item ) {
                        return item.msg_id == msg_id;
                } );

                if ( iter == context_.end() ) {
                        // TODO: this is bigass error
                        return;
                }

                // TODO: this returns error
                iter->tok.set_cb( buffer );
        }

        template < typename Visitor >
        void read_item( dynmx_reg_token tok, Visitor&& v )
        {
                transmit_servo< DYNMX_READ >( tok, v, tok.addr, tok.size );
        }

        template < typename Visitor >
        void set_item( dynmx_reg_token tok, Visitor&& v )
        {
                dynmx_buffer buff = tok.get_cb();
                transmit_servo< DYNMX_WRITE >( tok, v, tok.addr, buff );
        }

        template < dynmx_message_enum ID, typename Visitor, typename... Args >
        void transmit_servo( dynmx_reg_token tok, Visitor&& v, Args... args )
        {
                msg_id_counter_++;
                dynmx_crc crc;
                context_.emplace_back( msg_id_counter_, ID, tok );

                // error on full buffer

                v( servo_tag{},
                   msg_id_counter_,
                   master_dynmx_protocol::serialize(
                       dynmx_header_v, id_, master_dynmx_group::create< ID >( args... ), crc ) );
        }
};

}  // namespace schpin
