#include "node/events.h"
#include "node/lua_vm/extra_ser.h"

#include <schpin_lib/lua/pose.h>
#include <schpin_lib/lua/print.h>
#include <schpin_lib/lua/rosrequire.h>

#pragma once

namespace schpin
{

enum lua_call_enum
{
        CALL_ON_CMD
};

inline auto multi_joint_arg( std::string desc )
{
        return lua_multi_key_arg< std::string >{ "joint",
                                                 desc + "(either provide on joint, or list of "
                                                        "them)" };
}

struct slua_control_vm_def
{
        static auto get_def()
        {
                return std::make_tuple(
                    slua_print_proc,
                    slua_rosrequire{},
                    slua_proc(
                        "set_servo_config",
                        "sets configuration to servo",
                        slua_struct< lua_servo_config_event >(  // the nlohmann::json is pretty
                                                                // vague and that is not good :/
                            lua_key_arg< std::string >{ "joint", "joint to be configured" },
                            lua_key_arg< nlohmann::json >{ "vars", "set of config for servo" } ) ),
                    slua_func( "get_system_time",
                               "Returns actual system time in milliseconds",
                               slua_struct< get_system_time >(),
                               slua_result< std::size_t >() ),
                    slua_func( "get_leg_state",
                               "returns actual state of the leg",
                               slua_struct< get_leg_state_event >( lua_key_arg< std::string >{
                                   "leg", "name of the requested leg" } ),
                               slua_result< dof_array< em::angle > >{ "angles representing "
                                                                      "state" } ),
                    slua_func( "get_tip_pos",
                               "returns tip position of leg, when joints are in provided "
                               "configuration",
                               slua_struct< get_tip_pos_event >(
                                   lua_key_arg< std::string >{ "leg", "name of the requested leg" },
                                   lua_key_arg< dof_array< em::angle > >{ "angles",
                                                                          "a set of angles for "
                                                                          "given leg" } ),
                               slua_result< point< 3, robot_frame > >{ "point relative to robot "
                                                                       "frame" } ),
                    slua_proc(  //
                        "joint_goto",
                        "moves join to desired 'pos' in 'time'",
                        slua_struct< joint_goto_event >(  //
                            multi_joint_arg( "joints to be moved" ),
                            lua_key_arg< bangle >{ "pos", "position for the movement" },
                            lua_key_arg< btime >{ "time", "time for the movement in ms" } ) ),
                    slua_proc(  //
                        "joint_spin",
                        "spins joint for 'time' with 'velocity'",
                        slua_struct< joint_spin_event >(  //
                            multi_joint_arg( "joints to be spinned" ),
                            lua_key_arg< bangular_velocity >{ "velocity", "velocity of the spin" },
                            lua_key_arg< btime >{ "time", "how long does the spin take in ms" } ) ),
                    slua_proc(  //
                        "joint_spin_until",
                        "spins joint at 'velocity' until it arrives at 'pos'",
                        slua_struct< joint_spin_until_event >(  //
                            multi_joint_arg( "joints to be spinned" ),
                            lua_key_arg< bangular_velocity >{ "velocity",
                                                              "velocity of the movement" },
                            lua_key_arg< bangle >{ "pos",
                                                   "servo angle at which movement stops" } ) ),
                    slua_proc(  //
                        "send_read_pos",
                        "sends command for reading the position of joint",
                        slua_struct< send_read_pos_event >( multi_joint_arg( "joints that will be "
                                                                             "reported" ) ) ),
                    slua_func(  //
                        "sync",
                        "synchronizes execution of event-commands until all previous events "
                        "are "
                        "finished",
                        slua_struct< sync_event >( multi_joint_arg( "joints that should be "
                                                                    "synchronized" ) ),
                        slua_result< ctl_id >{ "id of the message" } ),
                    slua_func(  //
                        "get_joints",
                        "returns list of active joints",
                        slua_struct< get_joints_event >(),
                        slua_result< std::vector< std::string > >{ "list of joints" } ),
                    slua_proc(  //
                        "tip_goto",
                        "Moves tip of the `leg` to position 'target' in 'time'. 'steps'represents "
                        "number of iterations of the algorithm that solves the IK formula.",
                        slua_struct< tip_goto_event >(  //
                            lua_key_arg< std::string >{ "leg", "leg to be moved" },
                            lua_key_arg< btime >{ "time", "time of the movement" },
                            lua_key_arg< point< 3, world_frame > >{ "target",
                                                                    "target position of the tip of "
                                                                    "the leg" },
                            lua_opt_key_arg< std::size_t >{
                                "steps", "number of steps the IK solver", std::size_t{ 32 } } ) ),
                    slua_proc(  //
                        "relative_tip_goto",
                        "Moves tip of the `leg` to position `target` in `time`. The target "
                        "represents position relative to the positon of body",
                        slua_struct< relative_tip_goto_event >(  //
                            lua_key_arg< std::string >{ "leg", "leg to be moved" },
                            lua_key_arg< btime >{ "time", "time of the movement" },
                            lua_key_arg< point< 3, robot_frame > >{ "target",
                                                                    "target relative position of "
                                                                    "the tip of the leg" },
                            lua_opt_key_arg< std::size_t >{
                                "steps", "number of steps the IK solver", std::size_t{ 32 } } ) ),
                    slua_proc(  //
                        "tip_move",
                        "Moves tip of the `leg` in 'dir' for 'time'. 'steps' represents number of "
                        "iterations of the algorithm that solves the IK formula.",
                        slua_struct< tip_move_event >(  //
                            lua_key_arg< std::string >{ "leg", "leg to be moved" },
                            lua_key_arg< btime >{ "time", "time of the movement" },
                            lua_key_arg< vec< 3 > >{ "dir", "direction of the movement as vector" },
                            lua_opt_key_arg< std::size_t >{
                                "steps", "numer of steps the IK solver", std::size_t{ 32 } } ) ),
                    slua_proc(  //
                        "body_goto",
                        "Moves the body into a absolute coordinate specificed by pose in `time`.",
                        slua_struct< body_goto_event >(  //
                            slua_pose_args{ "point", "orientation", "The desired pose" },
                            lua_key_arg< btime >{ "time", "time of movement" },
                            lua_opt_key_arg< std::size_t >{
                                "steps", "number of steps the IK solver", std::size_t{ 32 } } ) ),
                    slua_proc(  //
                        "body_move",
                        "Moves body of the robot in direction 'dir' and rotates it by 'rot', by "
                        "moving all legs in opposite direction, this is done in 'time'. "
                        "'steps' represents number of iterations of the algorithm that solves the "
                        "IK formula.",
                        slua_struct< body_move_event >(  //
                            lua_key_arg< vec< 3 > >{ "dir", "direction of body movement" },
                            lua_key_arg< quaternion< robot_frame > >{ "rot",
                                                                      "rotation of the body" },
                            lua_key_arg< btime >{ "time", "time of the movement" },
                            lua_opt_key_arg< std::size_t >{
                                "steps", "number of steps the IK solver", std::size_t{ 32 } } ) ),
                    slua_proc(  //
                        "set_body_pose",
                        "Sets actual pose of the body of the robot relative to world frame",
                        slua_struct< set_body_pose_event >( slua_pose_args(  //
                            "point",
                            "orientation",
                            "The desired pose of the body in world frame" ) ) ),
                    /*slua_func( //
                            "get_body_pose",
                            "Gets actual pose of the robot body",
                            slua_struct< get_body_pose_event >(),
                            slua_result<pose<world_frame>>("actual pose of the robot body")),*/
                    slua_proc(  //
                        "print_state",
                        "Prints actual state of the robot",
                        slua_struct< print_state_event >() ),
                    slua_glob_var( "sync_cb", "Table of callbacks for messages", std::tuple<>{} ),
                    slua_glob_var( "msg_cb", "Callbacks for messages", std::tuple<>{} ),
                    slua_native_proc_call(  //
                        CALL_ON_CMD,
                        "on_cmd",
                        "Function is called each time msg is received from firmware",
                        lua_arg< fw_node_group_variant >{
                            "firmware message split to string" },  // msg_cb IS BIGASS HACK
                        R"EOF(
  if msg_cb[arg[0]] ~= nil then
    msg_cb[arg[0]](arg)
  end
	if arg[0] ~= "FINISHED" then
		return
	end
	msg_id = arg[1]
	if sync_cb[msg_id] ~= nil then
		sync_cb[msg_id]()
		sync_cb[msg_id] = nil
	end
		)EOF" ),
                    slua_native_proc(  //
                        "on_msg",
                        "executes callback when specific type of message arrives",
                        slua_native_args(  //
                            lua_key_arg< std::string >{ "key", "id of message" },
                            lua_key_arg< std::function< void() > >{ "cb", "callback" } ),
                        R"EOF(
    msg_cb[arg["key"]] = arg["cb"];
)EOF" ),
                    slua_native_proc(  //
                        "on_sync",
                        "synces and executes callback when that sync finishes",
                        slua_native_args(  //
                            lua_key_arg< std::vector< std::string > >{ "joints",
                                                                       "Joints which should be "
                                                                       "waited for" },
                            lua_key_arg< std::function< void() > >{ "callback",
                                                                    "callback to be executed" } ),
                        R"EOF(
	msg_id = sync{joints=arg["joints"]}
	sync_cb[msg_id] = arg["callback"]
		)EOF" ),
                    slua_native_proc(  //
                        "block_yield",
                        "blocks execution until all joints finish their task",
                        slua_native_args(  //
                            lua_key_arg< std::vector< std::string > >{ "joints",
                                                                       "Joints which should be "
                                                                       "waited for" } ),
                        R"EOF(
	finished = 0
	on_sync{joints=arg["joints"], callback=function()
		finished = 1
	end}
	while finished == 0 do
		coroutine.yield()
	end
		)EOF" ),
                    slua_native_class(  //
                        "Chain",
                        "Chain converts a coroutine into a iterable event loop. This makes it possible to chain up complex movements into one thing.",
                        slua_native_proc(  //
                            "Chain:new",
                            "Creates new chain, argument is a function",
                            lua_arg< std::function< void() > >{ "Coroutine to be sequenced." },
                            R"EOF(
            o = {fun = arg}
            setmetatable(o, self)
            self.__index = self
            return o
            )EOF" ),
                        slua_native_proc(  //
                            "Chain:exec",
                            "Executes the chain",
                            slua_native_args(),
                            R"EOF(
            co = coroutine.create(self.fun)
            while coroutine.resume(co) do
                coroutine.yield()
            end
            )EOF" ) ) );
        }
};

}  // namespace schpin
