#include "fw/serial.h"
#include "fw/servo/holder.h"
#include "node/lua_vm/extra_ser.h"

#include <ros/ros.h>
#include <schpin_lib/lua/repl.h>
#include <schpin_lib/lua/vm.h>
#include <schpin_lib_exp/config.h>

using namespace schpin;

struct slua_runner_def
{
        static auto get_def()
        {
                return std::make_tuple( slua_proc(  //
                    "joint_goto",
                    "Goto servo event",
                    slua_struct< lewan_goto_event >(
                        lua_key_arg< lewan_angle >{ "pos", "position in lewan angle" },
                        lua_key_arg< lewan_time >{ "time", "lewan time of the movement" } ) ) );
        }
};

class servo_runner
{
        serial_connection          ser_;
        servo_holder               servo_;
        slua_vm< slua_runner_def > vm_;
        slua_repl                  repl_;

      public:
        servo_runner( serial_connection ser )
          : ser_( ser )
          , servo_( lewan_servo_id::get< 13 >(), get_system_time() )
        {
        }

        // TODO: duplication from control_pc :/
        auto make_matcher()
        {
                return match(
                    [this]( control_tag, lewan_servo_id, auto ) {  //
                    },
                    [this]( activate_tag, servo_id id ) {
                            SCHPIN_INFO_LOG( "lewan_servo", "Got activation of: " << id );
                    },
                    [this]( servo_tag, s_ctl_id msg_id, reply r, const auto& m ) {
                            ser_.write( m );
                            if ( r == reply::NO ) {
                                    return;
                            }
                            receive_serial_msg( msg_id );
                    } );
        }
        void receive_serial_msg( s_ctl_id msg_id )
        {
                static constexpr std::size_t n = lewan_master_message::max_size;
                ser_.read< n >().match(
                    [&]( const em::static_vector< uint8_t, n >& data ) {
                            lewan_sequencer< lewan_master_message > sequencer;

                            bool load_fired = sequencer.load( data, [&]( lewan_master_message m ) {
                                    parse_message( msg_id, m );
                            } );

                            if ( !load_fired ) {
                                    SCHPIN_ERROR_LOG(
                                        "control",
                                        "servo not replying, got:" << convert_view< int >( data ) );
                            }
                    },
                    [&]( auto err ) {
                            SCHPIN_ERROR_LOG( "control", err );
                    } );
        }
        void parse_message( s_ctl_id msg_id, const lewan_master_message& m )
        {
                lewan_master_protocol::extract( m )
                    .bind_left( em::uncurry |
                                [&]( lewan_servo_id, lewan_master_submessage m ) {
                                        return lewan_master_group::extract( m );
                                } )
                    .match(
                        [&]( lewan_master_variant msg_var ) {
                                servo_.on_servo_receive( get_system_time(), msg_id, msg_var );
                        },
                        [&]( auto err ) {
                                SCHPIN_ERROR_LOG( "lewan_servo", err );
                        } );
        }

        void tick()
        {
                servo_.tick( get_system_time(), make_matcher() );
                servo_.poll( get_system_time(), 0, make_matcher() );
                std::string msg = repl_.prompt();
                vm_.load( msg ).optionally_log();
                vm_.full_exec( [&]( lua_State*, auto event ) {
                           handle_event( event );
                   } )
                    .optionally_log();
        }

        void handle_event( lewan_goto_event e )
        {
                servo_.on_event( ctl_id{ 0 }, get_system_time(), e, make_matcher() );
        }
};

struct servo_tester_config
{
        std::filesystem::path dev;

        static auto ros_param_get()
        {
                return std::make_tuple(  //
                    ros_opt_param< std::filesystem::path >(
                        "device", "Device to open for connection", "/dev/ttyUSB0" ) );
        }
};

int main( int argc, char* argv[] )
{

        ros::init( argc, argv, "servo_tester" );
        ros::AsyncSpinner spinner{ 2 };
        spinner.start();

        get_config< servo_tester_config >( "~" ).match(
            [&]( servo_tester_config conf ) {
                    serial_connection::make( conf.dev )
                        .match(
                            [&]( serial_connection p ) {
                                    servo_runner runnner{ p };

                                    while ( true ) {
                                            runnner.tick();
                                    }
                            },
                            [&]( auto e ) {
                                    std::cerr << e;
                            } );
            },
            [&]( auto e ) {
                    std::cerr << e;
            } );
}

