#include "octree/interfaces.h"
#include "octree/load.h"
#include "octree/occupancy.h"

#pragma once

namespace schpin
{
static constexpr uint16_t env_max_occupancy = 65535;
static constexpr uint16_t env_min_occupancy = 0;
static constexpr uint16_t env_prunning_diff = 1024;  // #random

class environment_oc_tree_content
  : public i_oc_content< environment_oc_tree_content, occupancy_token >
{
        uint16_t occupancy_  = env_max_occupancy / 2;
        bool     is_surface_ = false;

      public:
        environment_oc_tree_content() = default;

        environment_oc_tree_content( uint16_t occupancy, bool is_surface )
          : occupancy_( occupancy )
          , is_surface_( is_surface )
        {
        }

        uint16_t get_occupancy() const
        {
                return occupancy_;
        }

        void set_surface( bool val )
        {
                is_surface_ = val;
        }

        constexpr bool is_surface() const
        {
                return is_surface_;
        }

        std::size_t
        get_memory_usage( const oc_node< environment_oc_tree_content >& node ) const final
        {
                return sizeof( node );
        }

        bool prunnable( const oc_node< environment_oc_tree_content >& node ) const final
        {
                bool prunnable_occupancy = ( env_max_occupancy - occupancy_ ) < env_prunning_diff ||
                                           ( occupancy_ - env_min_occupancy ) < env_prunning_diff;

                bool prunnable_surface = em::none_of(
                    node.get_children(), [&]( const auto& child ) { return child->is_surface(); } );

                return prunnable_surface && prunnable_occupancy;
        }

        void child_based_update( const oc_node< environment_oc_tree_content >& node ) final
        {
                SCHPIN_ASSERT( node.size() != 0 );
                unsigned val = em::sum( node.get_children(),
                                        [&]( const oc_node< environment_oc_tree_content >& child ) {
                                                return child->get_occupancy();
                                        } );

                occupancy_ = uint16_t( val / node.size() );
        }

        void set_prunned( const oc_node< environment_oc_tree_content >& ) final
        {
                if ( occupancy_ > ( env_max_occupancy + env_min_occupancy ) / 2 ) {
                        occupancy_ = env_max_occupancy;
                } else {
                        occupancy_ = env_min_occupancy;
                }
        }

        void update( const occupancy_token& token ) final
        {
                switch ( token ) {
                        case occupancy_token::FULL:
                                occupancy_ = env_max_occupancy;
                                return;
                }
        }
};

struct is_surface_predicate
{
        constexpr bool operator()( const oc_node< environment_oc_tree_content >& node )
        {
                return node->is_surface();
        }
        template < typename Tag >
        constexpr bool operator()( const oc_proxy_node< environment_oc_tree_content, Tag >& node )
        {
                return this->operator()( *node );
        }
};

template < uint16_t Treshold >
struct env_occupancy_treshold_predicate
{
        constexpr bool operator()( const oc_node< environment_oc_tree_content >& node )
        {
                return node->get_occupancy() > Treshold;
        }
        template < typename Tag >
        constexpr bool operator()( const oc_proxy_node< environment_oc_tree_content, Tag >& node )
        {
                return this->operator()( *node );
        }
};

struct occupancy_collides_surface_predicate
{
        template < typename LhQuery, typename RhQuery >
        bool operator()( const LhQuery& lh, const RhQuery& rh )
        {
                bool is_occupied = occupancy_treshold_predicate{}( *lh );
                bool is_surface  = is_surface_predicate{}( *rh );

                return is_occupied && is_surface;
        }
};

template < uint16_t Treshold >
struct occupancy_collides_env_predicate
{
        template < typename LhQuery, typename RhQuery >
        constexpr bool operator()( const LhQuery& lh, const RhQuery& rh )
        {
                return occupancy_treshold_predicate{}( *lh ) &&
                       env_occupancy_treshold_predicate< Treshold >{}( *rh );
        }
};

inline bool operator==( const environment_oc_tree_content& lh,
                        const environment_oc_tree_content& rh )
{
        return lh.get_occupancy() == rh.get_occupancy() && lh.is_surface() == rh.is_surface();
}
inline bool operator!=( const environment_oc_tree_content& lh,
                        const environment_oc_tree_content& rh )
{
        return !( lh == rh );
}
inline void from_json( const nlohmann::json& j, environment_oc_tree_content& c )
{
        c = environment_oc_tree_content{ j.at( "occ" ).get< uint16_t >(),
                                         j.at( "is_surface" ).get< bool >() };
}
inline void to_json( nlohmann::json& j, const environment_oc_tree_content& c )
{
        j["occ"]        = c.get_occupancy();
        j["is_surface"] = c.is_surface();
}
}  // namespace schpin

template <>
struct std::hash< schpin::environment_oc_tree_content >
{
        std::size_t operator()( const schpin::environment_oc_tree_content& content ) const
        {
                return std::size_t( content.get_occupancy() ) ^ std::size_t( content.is_surface() );
        }
};
