#include "leg/data.h"
#include "leg/view.h"

#pragma once

namespace schpin
{

using leg_planner_error = error< struct leg_planner_error_tag >;

template < unsigned N >
class leg_planner
{
      public:
        using graph_search = d_star< leg_d_star_config< N, binary_heap > >;

        enum col_state
        {
                UNCHECKED = 0,
                COLLIDES,
                NEAR_COLLISION,
                NOT_COLLIDING
        };

      private:
        using state_cont = container< col_state, leg_vertex_id, std::vector >;

        graph_search sg_;
        state_cont   col_states_;

      public:
        explicit leg_planner( const leg_graph< N >& g )
          : sg_( g )
          , col_states_( g.get_vertexes().size() )
        {
        }

        leg_vertex_id find_closest_vertex( const leg_view< N >&           lview,
                                           const point< 3, world_frame >& position,
                                           leg_vertex_id                  lid,
                                           const environment&             env )
        {
                return lview.find_closest_vertex( position, lid, [&]( leg_vertex_id lid ) {
                        col_state vstate = make_vertex_check( lview, lid, env );
                        return vstate != COLLIDES;
                } );
        }

        col_state make_vertex_check( const leg_view< N >& lview,
                                     leg_vertex_id        vid,
                                     const environment&   env )
        {
                const leg_graph< N >& g = lview->get_graph();
                if ( col_states_[vid] != UNCHECKED ) {
                        return col_states_[vid];
                } else if ( lview.collides( g.get_vertex( vid )->angles, env ) ) {
                        col_states_[vid] = COLLIDES;
                } else if ( lview.near_surface( g.get_vertex( vid )->angles, env ) ) {
                        col_states_[vid] = NEAR_COLLISION;
                } else {
                        col_states_[vid] = NOT_COLLIDING;
                }
                return col_states_[vid];
        }

        em::either< chain< leg_edge_id, leg_vertex_id >, leg_planner_error >
        get_path( const leg_view< N >& leg_view,
                  leg_vertex_id        from,
                  leg_vertex_id        to,
                  const environment&   env )
        {
                const leg_graph< N >& g = leg_view->get_graph();
                sg_.set_task( g, from, to );
                chain< leg_edge_id, leg_vertex_id > p;
                chain< leg_edge_id, leg_vertex_id > prev;
                do {
                        sg_.compute_shortest_path( g, std::numeric_limits< std::size_t >::max() );
                        prev             = p;
                        p                = sg_.get_path( g );
                        bool is_bad_path = em::any_of( p.get_pins(), [&]( leg_vertex_id vid ) {
                                return sg_.get_extra_price( g, vid ) ==
                                       std::numeric_limits< leg_price >::max();
                        } );

                        if ( is_bad_path ) {
                                return { E( leg_planner_error )
                                         << "Failed to plan movement for "
                                            "leg from: "
                                         << from << " to: " << to
                                         << "; max price on path: " << em::view{ p.get_pins() } };
                        }
                        // THIS may cycle
                        for ( leg_vertex_id vid : p.get_pins() ) {
                                col_state vstate = make_vertex_check( leg_view, vid, env );
                                if ( vstate == COLLIDES ) {
                                        sg_.set_extra_price(
                                            g, vid, std::numeric_limits< leg_price >::max() );
                                }
                                if ( vstate == NEAR_COLLISION ) {
                                        // TODO: random constnat much
                                        sg_.set_extra_price( g, vid, leg_price{ 10 } );
                                }
                        }
                } while ( prev != p );
                return { p };
        }
};

extern template class leg_planner< 3 >;

}  // namespace schpin

template < unsigned N >
struct nlohmann::adl_serializer< schpin::leg_planner< N > >
{
        static void to_json( nlohmann::json& j, const schpin::leg_planner< N >& data )
        {
                j["graph"] = data.getGraph();
        }

        static schpin::leg_planner< N > from_json( const nlohmann::json& j )
        {
                return schpin::leg_planner< N >{ j.at( "graph" ).get< schpin::leg_graph< N > >() };
        }
};

template < unsigned N >
struct std::hash< schpin::leg_planner< N > >
{
        std::size_t operator()( const schpin::leg_planner< N >& planner )
        {
                // TODO maybe hash the graph search too?
                return std::hash< schpin::leg_graph< N > >()( planner.getGraph() );
        }
};
