#include "base/triangle.h"
#include "schpin_lib/either.h"

#include <schpin_lib/model/structure.h>
#include <string>
#include <utility>
#include <vector>

#pragma once

namespace schpin
{

using vec_vector_iterator = typename std::vector< vec< 3 > >::const_iterator;
template < typename Tag >
using point_set_iterator = typename std::set< point< 3, Tag > >::const_iterator;

template < typename TagT >
class structure_model_collision_query
  : i_collision_item< point_set_iterator< TagT >, vec_vector_iterator >
{
      public:
        using tag                      = TagT;
        using point_iterator           = point_set_iterator< tag >;
        using separation_axis_iterator = vec_vector_iterator;

      private:
        std::set< point< 3, tag > > points_;
        std::vector< vec< 3 > >     normals_;

      public:
        template < typename ModelTag >
        explicit structure_model_collision_query( const structure_model< ModelTag >& model,
                                                  pose< tag >                        offset )
        {
                for ( const triangle< 3, ModelTag >& tri : model.get_convex_hull() ) {
                        triangle< 3, tag > t = transform( tri, offset );

                        for ( const point< 3, tag >& p : t ) {
                                points_.insert( p );
                        }

                        normals_.push_back( normal_of( t ) );
                }
        }

        em::view< point_iterator > get_points() const final
        {
                return em::view{ points_ };
        }

        em::view< separation_axis_iterator > get_separation_axes() const final
        {
                return em::view{ normals_ };
        }

        em::view< separation_axis_iterator > get_intersection_axes() const final
        {
                return { normals_.end(), normals_.end() };
        }
};

template < typename RealTag, typename ModelTag >
class fast_structure_model_collision_query
  : i_fast_collision_item< structure_model_collision_query< RealTag >, 3 >
{
      public:
        using tag       = RealTag;
        using full_type = structure_model_collision_query< RealTag >;

      private:
        const structure_model< ModelTag >& model_;
        pose< RealTag >                    offset_;

      public:
        fast_structure_model_collision_query( const structure_model< ModelTag >& model,
                                              pose< RealTag >                    offset )
          : model_( model )
          , offset_( offset )
        {
        }
        point< 3, RealTag > get_min_sphere_position() const final
        {
                return transform( model_.get_center(), offset_ );
        }
        em::radius get_min_sphere_radius() const final
        {
                return model_.get_minball_radius();
        }
        structure_model_collision_query< RealTag > get_full() const final
        {
                return structure_model_collision_query< RealTag >{ model_, offset_ };
        }

        bool is_leaf() const
        {
                return false;
        }

        auto get_children() const
        {
                // TODO: this rather be lazy?
                return map_f_to_v(
                    model_.get_convex_hull(), [&]( const triangle< 3, ModelTag >& t ) {
                            return triangle_fast_collision_query< tag >( transform( t, offset_ ) );
                    } );
        }
};

}  // namespace schpin

