#include "planner/body.h"
#include "schpin_test/test_environment.h"

#include <gtest/gtest.h>
#include <tuple>

using namespace schpin;

extern test_data_archive ARCHIVE;

template < typename T >
class BodyPlannerTest : public ::testing::Test
{
      public:
        float               h = 0.07f;
        pose< world_frame > NEUTRAL_POSE{ point< 3, world_frame >{ 0, 0, h }, neutral_quat };
        pose< world_frame > SHIFTED_POSE{ point< 3, world_frame >{ 1, 0, h }, neutral_quat };
        pose< world_frame > CORNER_POSE{ point< 3, world_frame >{ 1, -1, h }, neutral_quat };
        body_planner_functions< T::mode, false > FUN;

        void check_path( bool                       should_be_valid,
                         const pose< world_frame >& from,
                         const pose< world_frame >& to )
        {
                const environment& env   = ARCHIVE.get_env( "column" );
                const robot&       robot = ARCHIVE.get_robot( "koke" );

                valid_path_res is_valid = FUN.is_valid_path( from, to, robot, env );
                EXPECT_EQ( should_be_valid, bool( is_valid ) ) << "from          : " << from << "\n"
                                                               << "to            : " << to << "\n"
                                                               << is_valid << std::endl;
        }
};

template < collision_mode ModeArg >
struct planner_conf
{
        static constexpr collision_mode mode = ModeArg;
};

using body_planner_test_types =
    ::testing::Types< planner_conf< collision_mode::FAST >, planner_conf< collision_mode::FULL > >;

TYPED_TEST_SUITE( BodyPlannerTest, body_planner_test_types );

TYPED_TEST( BodyPlannerTest, body_collides_with_environment )
{
        const environment& env   = ARCHIVE.get_env( "column" );
        const robot&       robot = ARCHIVE.get_robot( "koke" );

        EXPECT_TRUE( this->FUN.body_collides_with_environment(
            robot.get_body(), this->NEUTRAL_POSE, env ) );  // fails on collision in both modes
        EXPECT_FALSE(
            this->FUN.body_collides_with_environment( robot.get_body(), this->SHIFTED_POSE, env ) );
        EXPECT_FALSE(
            this->FUN.body_collides_with_environment( robot.get_body(), this->CORNER_POSE, env ) );
}

TYPED_TEST( BodyPlannerTest, all_legs_touches_with_environment )
{
        const environment& env   = ARCHIVE.get_env( "column" );
        const robot&       robot = ARCHIVE.get_robot( "koke" );

        EXPECT_TRUE( this->FUN.all_legs_touches_with_environment(
            robot.get_legs(), this->CORNER_POSE, env ) );
        EXPECT_TRUE( this->FUN.all_legs_touches_with_environment(
            robot.get_legs(), this->NEUTRAL_POSE, env ) );
        EXPECT_TRUE( this->FUN.all_legs_touches_with_environment(
            robot.get_legs(), this->SHIFTED_POSE, env ) );
}

TYPED_TEST( BodyPlannerTest, isValidPath )
{
        std::vector< std::tuple< bool, pose< world_frame >, pose< world_frame > > > tests{
            { false, this->NEUTRAL_POSE, this->SHIFTED_POSE },
            { false,
              this->NEUTRAL_POSE,
              pose{ point< 3, world_frame >{ 1, 1, this->h }, neutral_quat } },
            { false,
              this->NEUTRAL_POSE,
              pose{ point< 3, world_frame >{ -5, 0, this->h }, neutral_quat } },
            { true, this->SHIFTED_POSE, this->CORNER_POSE },
            { true,
              this->CORNER_POSE,
              pose{ point< 3, world_frame >{ 0.3, 0.3, this->h }, neutral_quat } } };

        for ( auto [should_be_valid, from, to] : tests ) {
                this->check_path( should_be_valid, from, to );
                this->check_path( should_be_valid, to, from );
        }
}

TYPED_TEST( BodyPlannerTest, simple )
{
        static constexpr collision_mode mode = TypeParam::mode;

        const environment& env   = ARCHIVE.get_env( "column" );
        const robot&       robot = ARCHIVE.get_robot( "koke" );

        planner_context context{
            this->SHIFTED_POSE, std::make_optional( this->CORNER_POSE ), env.copy(), robot.copy() };

        body_planner< mode, false > planner{ context };

        body_planner_lambda_visitor visitor{ []( const auto& ) {  //
                return opt_error< body_path_failure_event >{};
        } };

        std::size_t max_steps = 50;
        std::size_t step      = *em::find_if( em::range( max_steps ), [&]( std::size_t ) -> bool {
                planner.event( make_vertexes_event{ 1 }, context, visitor ).optionally( []( auto ) {
                } );
                planner.event( find_body_path_event{}, context, visitor ).optionally( []( auto ) {
                } );

                return planner.has_plan();
        } );

        EXPECT_NE( max_steps, step );
}
